''' Represents DSO name and object panel (including entry of new DSO info)
'''

import os
import math

from kivy.app import App
from kivy.properties import StringProperty, BooleanProperty, ObjectProperty, DictProperty
from kivy.lang import Builder
from kivy.uix.button import Button
from kivy.uix.boxlayout import BoxLayout
from kivy.logger import Logger
from kivy.clock import Clock
from kivy.metrics import dp
from kivy.core.clipboard import Clipboard

from jocular.component import Component
from jocular.widgets import JWidget, JPopup
from jocular.RA_and_Dec import RA, Dec

Builder.load_string(
    '''

<MatchButton>:
    size_hint: 1, None
    markup: True
    text_size: self.size
    padding: dp(10), dp(1)
    halign: 'left'
    font_size: '14sp'
    color: app.lowlight_color
    height: dp(18)

<DSO>:
    size_hint: None, None
    orientation: 'vertical'
    padding: dp(5), dp(2)
    size: app.gui.width / 3, app.gui.height / 3
    y: .67 * app.gui.height
    x: 0 if root.show_DSO else -self.width

    Button:
        size_hint: 1, None
        font_size: '20sp'
        text: '[b]{:}[/b]'.format(root.Name) if root.Name else 'DSO name'
        on_press: root.edit()
        height: dp(30)
        markup: True
        text_size: self.size
        padding: dp(3), dp(1)
        halign: 'left'
        color: app.lowlight_color
        background_color: 0, 0, 0, 0

    ParamValue:
        param: 'Type'
        value: root.otypes.get(root.OT, root.OT)
        callback: root.edit
    ParamValue:
        param: 'Con'
        value: root.Con
        callback: root.edit
    ParamValue:
        param: 'RA'
        value: root.RA
        callback: root.paste_RA
        #callback: Clipboard.copy(self.value)
    ParamValue:
        param: 'Dec'
        value: root.Dec
        #callback: Clipboard.copy(self.value)
        callback: root.paste_Dec
    ParamValue:
        param: 'Mag'
        value: root.Mag 
        callback: root.edit
    ParamValue:
        param: 'Diam'
        value: root.Diam
        callback: root.edit
    ParamValue:
        param: 'Info'
        value: root.Other
        callback: root.edit

    Label:
        size_hint: 1, 1

<ParamLabel@Label>:
    size_hint: None, None
    size: dp(100), dp(26)
    text_size: self.size
    padding: dp(5), dp(1)
    markup: True
    halign: 'right'
    valign: 'center'
    color: app.lowlight_color
    font_size: app.form_font_size

<WhiteParamLabel@ParamLabel>:
    canvas.before:
        Color:
            rgba: .8, .8, .8, 1
        Rectangle:
            pos: self.pos
            size: self.size
    color: 0, 0, 0, 1

<ParamValue2@TextInput>:
    size_hint: None, None
    size: dp(50), dp(26)
    halign: 'right'
    valign: 'center'
    background_color: .8, .8, .8, 1
    font_size: app.form_font_size

<MyBoxLayout>:
    size_hint: 1, None
    height: dp(24)

<DSOInfo>:
    orientation: 'vertical'
    size_hint: None, None
    size: dp(360), dp(630)
    spacing: dp(2) # 3 # dp(3)
    matches: _matches
    dso_input: _dso

    Label:
        size_hint: 1, None
        height: dp(20)
        color: app.lowlight_color
        background_color: 0, 0, 0, 0
        text: 'Type DSO name below, choose from list and "select"'
    Label:
        size_hint: 1, None
        height: dp(20)
        color: app.lowlight_color
        background_color: 0, 0, 0, 0
        text: 'Or add/edit object information and "update catalogue"'

    TextInput:
        unfocus_on_touch: False
        id: _dso
        font_size: '18sp'
        text: root.dso.Name
        size_hint: 1, None
        valign: 'center'
        height: dp(30)
        on_text: root.lookup_dso(self.text)
        background_color: .8, .8, .8, 1

    Label:
        size_hint: 1, 1

    BoxLayout:
        canvas.before:
            Color:
                rgba: .1, .1, .1, 1
            Rectangle:
                pos: self.pos
                size: self.size
        id: _matches
        size_hint: 1, None
        height: dp(200)
        orientation: 'vertical'

    Label:
        size_hint: 1, 1

    MyBoxLayout:
        ParamLabel:
            text: '[b]Name[/b]'
        ParamValue2:
            text: root.Name
            on_text: root.check(self, 'Name')
            size_hint: 1, 1
            halign: 'left'

    MyBoxLayout:
        ParamLabel:
            text: '[b]Object type[/b]'
        ParamValue2:
            text: root.OT
            on_text: root.check(self, 'OT')
        ParamLabel:
            size_hint: .3, 1
            text: '1-3 char code'
            halign: 'left'

    MyBoxLayout:
        ParamLabel:
            text: '[b]Constellation[/b]'
        ParamValue2:
            on_text: root.check(self, 'Con')
            text: root.Con
        ParamLabel:
            size_hint: .3, 1
            text: '3 char code'
            halign: 'left'

    MyBoxLayout:
        ParamLabel:
            text: '[b]RA[/b]'
        ParamValue2:
            text: root.RA_h
            on_text: root.check(self, 'RA_h')
        WhiteParamLabel:
            width: dp(15)
            text: 'h'
        ParamValue2:
            on_text: root.check(self, 'RA_m')
            text: root.RA_m
        WhiteParamLabel:
            width: dp(15)
            text: "m"
        ParamValue2:
            text: root.RA_s
            on_text: root.check(self, 'RA_s')
        WhiteParamLabel:
            width: dp(15)
            text: 's'

    MyBoxLayout:
        ParamLabel:
            text: '[b]Dec[/b]'
        ParamValue2:
            text: root.Dec_d
            on_text: root.check(self, 'Dec_d')
        WhiteParamLabel:
            width: dp(15)
            text: '\u00b0'
        ParamValue2:
            text: root.Dec_m
            on_text: root.check(self, 'Dec_m')
        WhiteParamLabel:
            width: dp(15)
            text: "'"
        ParamValue2:
            text: root.Dec_s
            on_text: root.check(self, 'Dec_s')
        WhiteParamLabel:
            width: dp(15)
            text: '"'

    MyBoxLayout:
        ParamLabel:
            text: '[b]Magnitude[/b]'
        ParamValue2:
            text: root.Mag
            on_text: root.check(self, 'Mag')
 
    MyBoxLayout:
        ParamLabel:
            text: '[b]Diameter[/b]'
        ParamValue2:
            text: root.Diam
            on_text: root.check(self, 'Diam')
        ParamLabel:
            size_hint: .3, 1
            text: 'arcmins'
            halign: 'left'

    MyBoxLayout:
        ParamLabel:
            text: '[b]Other[/b]'
        ParamValue2:
            text: root.Other
            on_text: root.Other = self.text
            size_hint: 1, 1
            halign: 'left'

    MyBoxLayout:
        height: dp(30)
        Label:
            size_int: .67, .8
        Button:
            text: 'Clear fields'
            size_hint: .33, .8
            on_press: root.clear_DSO()

    Label:
        size_hint: 1, 1

    MyBoxLayout:
        height: dp(35)
        Button:
            text: 'Add name'
            size_hint: .3, .8
            on_press: root.done()
        Button:
            text: 'Update catalogue'
            size_hint: .4, .8
            on_press: root.add_to_catalogue()
        Button:
            text: 'Cancel'
            size_hint: .3, .8
            on_press: root.dso.popup.dismiss()

    ParamLabel:
        size_hint: 1, None
        height: dp(35)
        text: root.message
        halign: 'center'

'''
)


class MyBoxLayout(BoxLayout):
    pass

class MatchButton(Button, JWidget):
    pass

def split_name_and_OT(s):
    ''' names also contain object type in format name/type; need to handle 
        the few names that also contain a /
    '''
    parts = s.split('/')
    if len(parts) == 1:
        return parts, ''
    if len(parts) == 2:
        return parts[0], parts[1]
    if len(parts) > 2:
        return '/'.join(parts[:-1]), parts[-1]


class DSOInfo(BoxLayout):

    Name = StringProperty('')
    OT = StringProperty('')
    Con = StringProperty('')
    RA_h = StringProperty('')
    RA_m = StringProperty('')
    RA_s = StringProperty('')
    Dec_d = StringProperty('')
    Dec_m = StringProperty('')
    Dec_s = StringProperty('')
    Mag = StringProperty('')
    Diam = StringProperty('')
    Other = StringProperty('')
    message = StringProperty('')
    matches = ObjectProperty(None)
    all_valid = BooleanProperty(False)
    dso_input = ObjectProperty(None)

    def __init__(self, dso, **kwargs):
        self.dso = dso

        # fill in current values
        for p in ['Name', 'Con', 'OT', 'Mag', 'Diam', 'Other']:
            setattr(self, p, getattr(dso, p))

        # deal with RA/Dec
        try:
            self.RA_h, self.RA_m, self.RA_s = dso.RA.replace('h', '').split()
            self.Dec_d, self.Dec_m, self.Dec_s = dso.Dec.replace('\u00b0', '').split()
        except:
            pass

        super().__init__(**kwargs)
        self.max_matches = 12
        self.match_buttons = [
            MatchButton(on_press=self.choose_dso) for i in range(self.max_matches)
        ]
        for b in self.match_buttons:
            self.matches.add_widget(b)

        Clock.schedule_once(self.set_focus, 0)

    def set_focus(self, dt):
        self.dso_input.focus = True

    def on_Mag(self, *args):
        if self.Mag == 'nan':
            self.Mag = ''

    def on_Diam(self, *args):
        if self.Diam == 'nan':
            self.Diam = ''

    def check_value(self, name, value):
        try:
            if name == 'Name':
                ok = len(value) > 0
                mess = 'must supply a name'
            elif name == 'Mag':
                ok = self.in_range(value, -20, 40, nonempty=False)
                mess = 'magnitude out of range'
            elif name == 'OT':
                ok = len(value) in [1, 2, 3]
                mess = 'object type must be 1 to 3-char code'
            elif name == 'Con':
                ok = len(value) == 3
                mess = 'constellation must be 3-char code'
            elif name == 'RA_h':
                ok = self.in_range(value, 0, 23)
                mess = 'RA hours must be 0-23'
            elif name == 'RA_m':
                ok = self.in_range(value, 0, 60)
                mess = 'RA mins must be 0-60'
            elif name == 'RA_s':
                ok = self.in_range(value, 0, 60)
                mess = 'RA secs must be 0-60'
            elif name == 'Dec_d':
                ok = self.in_range(value, -90, 90)
                mess = 'Dec degrees must be -90 to +90'
            elif name == 'Dec_m':
                ok = self.in_range(value, 0, 60)
                mess = 'Dec mins must be 0-60'
            elif name == 'Dec_s':
                ok = self.in_range(value, 0, 60)
                mess = 'Dec secs must be 0-60'
            elif name == 'Diam':
                ok = self.in_range(value, 0, 9999, nonempty=False)
                mess = 'Diameter must be a positive number'
            elif name == 'Other':
                ok = True
                mess = ''
        except Exception:
            ok = False
            mess = 'validation issue'

        if ok:
            mess = None

        return mess

    def check(self, field, name):
        value = field.text.strip()
        mess = self.check_value(name, value)
        setattr(self, name, value)
        if mess is None:
            self.message = ''
            field.foreground_color = 0, 0, 0, 1
        else:
            self.message = mess
            field.foreground_color = 1, 0, 0, 1

    def in_range(self, val, lo, hi, nonempty=True):
        if nonempty and not val:
            return False
        if not nonempty and val == 'nan':
            return True
        if not nonempty and not val:
            return True
        try:
            v = float(val)
            return v <= hi and v >= lo
        except:
            return False

    def done(self, *args):
        # validate properties and form into suitable dict for DSO to display

        props = {}
        for p in ['OT', 'Con', 'Diam', 'Mag', 'Other']:
            value = getattr(self, p)
            if self.check_value(p, value) is None:
                props[p] = value
            else:
                props[p] = ''

        props['Name'] = self.Name

        if (
            self.check_value('RA_h', self.RA_h) is None
            and self.check_value('RA_m', self.RA_m) is None
            and self.check_value('RA_s', self.RA_s) is None
        ):
            props['RA'] = '{:2s}h {:2s} {:2s}'.format(self.RA_h, self.RA_m, self.RA_s)
        else:
            props['RA'] = ''

        if (
            self.check_value('Dec_d', self.Dec_d) is None
            and self.check_value('Dec_m', self.Dec_m) is None
            and self.check_value('Dec_s', self.Dec_s) is None
        ):
            props['Dec'] = '{:s}\u00b0 {:2s} {:2s}'.format(
                self.Dec_d, self.Dec_m, self.Dec_s
            )
        else:
            props['Dec'] = ''

        self.dso.selected(props)

    def choose_dso(self, but):
        if len(but.text.strip()) > 0:
            try:
                ot, name = but.text.split(':')
                self.show_DSO(name.strip(), ot.strip())
            except:
                pass

    def show_DSO(self, name, ot):
        self.Name = name
        details = Component.get('ObservingList').lookup_details(
            '{:}/{:}'.format(name, ot)
        )
        self.OT = details.get('OT', '')
        self.Con = details.get('Con', '')
        self.Mag = str(details.get('Mag', ''))
        self.Diam = str(details.get('Diam', ''))
        self.Other = str(details.get('Other', ''))
        ra1, ra2, ra3 = str(RA(details.get('RA', ''))).split(' ')
        self.RA_h = ra1.strip()[:-1]
        self.RA_m = ra2.strip()
        self.RA_s = ra3.strip()
        dec1, dec2, dec3 = str(Dec(details.get('Dec', ''))).split(' ')
        self.Dec_d = dec1.strip()[:-1]
        self.Dec_m = dec2.strip()
        self.Dec_s = dec3.strip()



    def clear_DSO(self):
        for p in ['Name', 'OT', 'Con', 'RA_h', 'RA_m', 'RA_s', 'Dec_d', 'Dec_m', 'Dec_s', 'Mag', 'Diam', 'Other']:
            setattr(self, p, '')

    def add_to_catalogue(self):

        # check that data is all valid
        for p in [
            'Name',
            'OT',
            'Con',
            'RA_h',
            'RA_m',
            'RA_s',
            'Dec_d',
            'Dec_m',
            'Dec_s',
            'Mag',
            'Diam',
            'Other',
        ]:
            resp = self.check_value(p, getattr(self, p))
            if resp is not None:
                self.message = resp
                return

        # check if catalogue exists; if not, create it and write header
        # append row of data
        try:
            path = App.get_running_app().get_path('catalogues')
            obj_file = os.path.join(path, 'user_objects.csv')
            if not os.path.exists(obj_file):
                with open(obj_file, 'w') as f:
                    f.write('Name,RA,Dec,Con,OT,Mag,Diam,Other\n')
        except Exception as e:
            self.message = 'Problem creating user objects file'
            Logger.error('DSO: problem creating objects list ({:})'.format(e))
            return

        try:
            ra = 15 * float(self.RA_h) + float(self.RA_m) / 4 + float(self.RA_s) / 240
            if float(self.Dec_d) < 0:
                dec = (
                    float(self.Dec_d)
                    - float(self.Dec_m) / 60
                    - float(self.Dec_s) / 3600
                )
            else:
                dec = (
                    float(self.Dec_d)
                    + float(self.Dec_m) / 60
                    + float(self.Dec_s) / 3600
                )
        except Exception as e:
            self.message = 'Problem converting RA or Dec'
            Logger.error('DSO: coverting RA or Dec ({:})'.format(e))
            return

        try:
            Name = self.Name.strip()
            OT = self.OT.strip().upper()
            Con = self.Con.strip().upper()
            Mag = self.Mag.strip()
            if len(Mag) == 0:
                Mag = math.nan
            Diam = self.Diam.strip()
            if len(Diam) == 0:
                Diam = math.nan
            Other = self.Other.strip().replace(',', ';')
            with open(obj_file, 'a') as f:
                f.write(
                    '{:},{:.6f},{:.6f},{:},{:},{:},{:},{:}\n'.format(
                        Name, ra, dec, Con, OT, Mag, Diam, Other
                    )
                )
        except Exception as e:
            self.message = 'Problem writing to user objects file'
            Logger.error('DSO: writing to user objects file ({:})'.format(e))
            return

        # update objects listing (overwriting, if duplicate exists)
        ol = Component.get('ObservingList')
        try:
            ol.objects['{:}/{:}'.format(Name, OT)] = {
                'Name': Name,
                'RA': ra,
                'Dec': dec,
                'Con': Con,
                'OT': OT,
                'Obs': 0,
                'Added': '',
                'List': 'N',
                'Other': Other,
                'Notes': '',
                'Mag': float(Mag),
                'Diam': float(Diam),
            }
            ol.compute_transits()
            ol.update_status()

        except Exception as e:
            self.message = 'Problem updating objects list'
            Logger.error('DSO: problem updating objects list ({:})'.format(e))
            return

        self.message = 'Written to catalogue!'



    def lookup_dso(self, name):
        ''' User has changed name field so look it up.
        '''

        try:
            self.clear_DSO()
            for b in self.match_buttons:
                b.text = ''
            name = name.strip()
            if name:
                matches = sorted(
                    Component.get('ObservingList').lookup(
                        name, max_matches=self.max_matches
                    )
                )
                if len(matches) == 1:
                    name, ot = split_name_and_OT(matches[0])
                    self.show_DSO(name.strip(), ot.strip())
                else:
                    for i, m in enumerate(matches):
                        nm, ot = split_name_and_OT(m)
                        self.match_buttons[i].text = '{:}: {:}'.format(ot, nm)
        except Exception as e:
            Logger.warn('DSO: problem looking up DSO {:} ({:})'.format(name, e))


class DSO(BoxLayout, Component):

    Name = StringProperty('')
    Con = StringProperty('')
    RA = StringProperty('')
    Dec = StringProperty('')
    OT = StringProperty('')
    Mag = StringProperty('')
    Diam = StringProperty('')
    Other = StringProperty('')
    otypes = DictProperty({})

    props = ['Name', 'Con', 'OT', 'RA', 'Dec', 'Mag', 'Diam', 'Other']

    show_DSO = BooleanProperty(False)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        otypes = Component.get('Catalogues').get_object_types()
        self.otypes = {k: v['name'] for k, v in otypes.items()}
        App.get_running_app().gui.add_widget(self, index=8)
        Logger.info('DSO: initialised')

    def on_Mag(self, *args):
        if self.Mag == 'nan':
            self.Mag = ''

    def on_Diam(self, *args):
        if self.Diam == 'nan':
            self.Diam = ''
        else:
            try:
                self.Diam = '{:.2f}'.format(float(self.Diam))
            except Exception:
                self.Diam = ''

    def on_new_object(self, settings=None):
        ''' Called at start and whenever user selects new object. Also called
            when user selects a row from the DSO table.
        '''
        if settings is None:
            self.settings = Component.get('Metadata').get({'Name', 'OT'})
            self._new_object()
        else:
            ''' Ask user for confirmation that they wish to rename the object
            '''
            self.settings = settings
            if self.Name:
                self.popup = JPopup(
                    title='Change name to {:}?'.format(settings['Name']),
                    actions={
                        'Cancel': self.cancel_name_change,
                        'Yes': self.confirmed_name_change,
                    },
                )
                self.popup.open()
            else:
                self.confirmed_name_change()

    def confirmed_name_change(self, *args):
        self.changed = not Component.get('Stacker').is_empty()
        self._new_object()

    def cancel_name_change(self, *args):
        pass

    def _new_object(self, *args):
        ''' Helper method called after normal new object selection or confirmation
            of change of name
        '''
        settings = self.settings

        if 'Name' in settings and 'OT' in settings:
            #  translate a few cases to new OTs
            nm = settings['Name']
            ot = settings['OT'].upper()
            # deal with some legacies from v1
            if ot == 'G+':
                if nm.startswith('Arp') or nm.startswith('VV') or nm.startswith('AM '):
                    ot = 'PG'
                elif (
                    nm.startswith('Hick')
                    or nm.startswith('PCG')
                    or nm.startswith('SHK')
                ):
                    ot = 'CG'
            lookup_settings = Component.get('ObservingList').lookup_name(
                '{:}/{:}'.format(nm, ot)
            )
            if lookup_settings is not None:
                settings = lookup_settings

        # now extract all
        for p in self.props:
            setattr(self, p, str(settings.get(p, '')))
        self.saved_props = {p: getattr(self, p) for p in self.props}

    def on_save_object(self):
        Component.get('Metadata').set(
            {'Name': self.Name.strip(), 'OT': self.OT, 'Con': self.Con}
        )

    def current_object_coordinates(self):
        if self.RA and self.Dec:
            return (float(RA(self.RA)), float(Dec(self.Dec)))
        return None, None

    def paste_RA(self, *args):
        Clipboard.copy(self.RA)

    def paste_Dec(self, *args):
        Clipboard.copy(self.Dec)

    def edit(self, *args):
        content = DSOInfo(self)
        self.popup = JPopup(
            title='Select, add or edit DSO', content=content, posn='top-left'
        )
        self.popup.open()

    def selected(self, new_props, *args):
        for p in self.props:
            if p in new_props:
                setattr(self, p, new_props[p])
            else:
                setattr(self, p, '')
        self.changed = (
            self.saved_props != {p: getattr(self, p) for p in self.props}
            and not Component.get('Stacker').is_empty()
        )
        self.popup.dismiss()

    # commented this while seeking weird touch down bug
    def on_touch_down(self, touch):
        # print(' TD in DSO')
        if (
            self.collide_point(*touch.pos)
            and touch.pos[0] < dp(100)
            and App.get_running_app().showing == 'main'
        ):
            return super().on_touch_down(touch)
        return False



''' Handles DSO table and observing list management, including alt-az/transit
    computations.
'''

import json
import time
import math
import numpy as np
from datetime import datetime, timedelta
from collections import Counter

from kivy.app import App
from kivy.clock import Clock
from kivy.core.window import Window
from kivy.properties import ConfigParserProperty
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.slider import Slider
from kivy.uix.label import Label
from kivy.metrics import dp
from kivy.logger import Logger

from jocular.component import Component
from jocular.RA_and_Dec import RA, Dec
from jocular.table import Table, CButton, TableLabel
from jocular.calcs import local_sidereal_time, sun_altitude


def ToHM(x):
    if math.isnan(x):
        return ''
    h = int(x)
    m = (x - h) * 60
    return '{:2.0f}h{:02.0f}'.format(h, m).strip()


def fmt_diam(d):
    if d < 1:
        # ds = 60
        return '{:4.1f}"'.format(d * 60)
    elif d < 10:
        return "{:4.2f}'".format(d)
    elif d < 100:
        return "{:4.1f}'".format(d)
    else:
        dg = int(d / 60)
        return "{:.0f}\u00b0{:2.0f}'".format(dg, (d - dg) / 60.0)


def quadrant(x):
    # terrestrial direction
    qmap = ['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']
    quad = int((x + 22.5) // 45)
    if quad == 8:
        quad = 0
    return qmap[quad]


class ObservingList(Component):

    latitude = ConfigParserProperty(
        50, 'Observatory', 'latitude', 'app', val_type=float
    )
    longitude = ConfigParserProperty(
        -2, 'Observatory', 'longitude', 'app', val_type=float
    )

    def __init__(self):
        super().__init__()
        self.app = App.get_running_app()

        # time zone offset
        ts = time.time()
        self.utc_offset = (
            datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)
        ).total_seconds() / 3600.0

        Clock.schedule_once(self.load, 0)

    def on_close(self):
        self.save_observing_list()
        self.save_notes()

    def save_observing_list(self, *args):
        try:
            ol = {
                v['Name']: v['Added'] for v in self.objects.values() if v['Added'] != ''
            }
            with open(self.app.get_path('observing_list.json'), 'w') as f:
                json.dump(ol, f, indent=1)
        except Exception as e:
            Logger.error('ObservingList: problem saving observing list ({:})'.format(e))
            self.warn('problem saving list')

    def save_notes(self, *args):
        try:
            ol = {
                v['Name']: v['Notes']
                for v in self.objects.values()
                if v['Notes'].strip() != ''
            }
            with open(self.app.get_path('observing_notes.json'), 'w') as f:
                json.dump(ol, f, indent=1)
        except Exception as e:
            Logger.error(
                'ObservingList: problem saving observing notes ({:})'.format(e)
            )
            self.warn('problem saving notes')

    def compute_altaz(self, current_hours_offset=0):

        t = datetime.now() + timedelta(seconds=3600 * current_hours_offset)
        rads = math.pi / 180

        if hasattr(self, 'last_time_changed'):
            tdiff = t - self.last_time_changed
            if tdiff.seconds < 5:
                return
        self.last_time_changed = t

        lst = local_sidereal_time(t, self.longitude)
        lat = self.latitude * rads
        ras = np.array([v['RA'] for v in self.objects.values()])
        decs = np.array([v['Dec'] for v in self.objects.values()]) * rads

        sinlat, coslat = np.sin(lat), np.cos(lat)
        H = (-self.utc_offset * 15 + lst - ras) * rads
        sinH, cosH = np.sin(H), np.cos(H)
        az = 180 + (np.arctan2(sinH, cosH * sinlat - np.tan(decs) * coslat) / rads)
        alt = np.arcsin(sinlat * np.sin(decs) + coslat * np.cos(decs) * cosH) / rads

        for v, _alt, _az in zip(self.objects.values(), alt, az):
            v['Az'] = int(_az)
            v['Alt'] = int(_alt)
            v['Quadrant'] = quadrant(_az)

    def compute_transits(self):
        # from ch 15 of Meeus

        t0 = time.time()

        ras = np.array([v['RA'] for v in self.objects.values()])
        decs = np.array([v['Dec'] for v in self.objects.values()])

        #  apparent sidereal time at Greenwich at 0hUT
        t = datetime.now()
        gst_0 = local_sidereal_time(datetime(t.year, t.month, t.day), 0)

        max_alts = 90 - self.latitude + decs
        max_alts[max_alts > 90] = 180 - max_alts[max_alts > 90]

        # transit time: slight diff from Meeus since longitude negative from W here
        transiting = np.abs(decs) < (90 - self.latitude)
        m0 = (ras - self.longitude - gst_0 + self.utc_offset * 15) / 15
        m0[m0 > 24] = m0[m0 > 24] - 24
        m0[m0 < 0] = m0[m0 < 0] + 24

        # 80 ms or so
        for v, _max_alt, _transits, _m in zip(
            self.objects.values(), max_alts, transiting, m0
        ):
            v['MaxAlt'] = int(_max_alt)
            v['Transit'] = float(_m) if _transits else math.nan

        Logger.info(
            'ObservingList: computed transits for {:} objects in {:.0f} ms'.format(
                len(self.objects), 1000 * (time.time() - t0)
            )
        )


    def load(self, dt):

        # DSOs
        try:
            self.objects = Component.get('Catalogues').get_basic_dsos()
        except Exception as e:
            Logger.exception('ObservingList: problem loading DSOs ({:})'.format(e))
            self.warn('problem loading DSOs')
            self.objects = {}

        # load observing list (maps names to date added)
        try:
            with open(self.app.get_path('observing_list.json'), 'r') as f:
                observing_list = json.load(f)
        except:
            observing_list = {}
        Logger.info('ObservingList: {:} on observing list'.format(len(observing_list)))

        # load observing notes
        try:
            with open(self.app.get_path('observing_notes.json'), 'r') as f:
                observing_notes = json.load(f)
        except:
            observing_notes = {}
        Logger.info('ObservingList: {:} observing notes'.format(len(observing_notes)))

        # load previous observations if necc and count previous
        try:
            obs = Component.get('Observations').get_observations()
            previous = dict(
                Counter([v['Name'].lower() for v in obs.values() if 'Name' in v])
            )
        except Exception as e:
            Logger.exception(
                'ObservingList: problem loading previous DSOs ({:})'.format(e)
            )
            previous = {}
        Logger.info(
            'ObservingList: {:} unique previous observations found'.format(
                len(previous)
            )
        )

        # ~ 180 ms
        for v in self.objects.values():
            name = v['Name']
            v['Obs'] = previous.get(name.lower(), 0)
            v['Added'] = observing_list.get(name, '')
            v['List'] = 'Y' if name in observing_list else 'N'
            v['Notes'] = observing_notes.get(name, '')
            v['Other'] = v.get('Other', '')

        try:
            self.compute_transits()
        except Exception as e:
            Logger.error('ObservingList: problem computing transits {:}'.format(e))

        self.update_status()

    def build(self):

        if not hasattr(self, 'objects'):
            self.load()

        t0 = time.time()

        cols = {
            'Name': {
                'w': 180,
                'align': 'left',
                'sort': {'catalog': ''},
                'action': self.new_from_list,
            },
            'OT': {'w': 45},
            'Con': {'w': 45},
            'RA': {
                'w': 80,
                'align': 'right',
                'type': float,
                'val_fn': lambda x: x * 15,
                'display_fn': lambda x: str(RA(x)),
            },
            'Dec': {
                'w': 85,
                'align': 'right',
                'type': float,
                'display_fn': lambda x: str(Dec(x)),
            },
            'Loc': {'w': 40, 'align': 'center', 'field': 'Quadrant'},
            'Az': {'w': 40, 'align': 'center', 'type': int},
            'Alt': {'w': 40, 'align': 'center', 'type': int},
            'Max': {'w': 40, 'align': 'center', 'type': int, 'field': 'MaxAlt'},
            'Transit': {'w': 60, 'align': 'right', 'type': float, 'display_fn': ToHM},
            'Mag': {'w': 50, 'align': 'right', 'type': float},
            'Diam': {'w': 50, 'align': 'right', 'type': float, 'display_fn': fmt_diam},
            'Obs': {'w': 40, 'type': int, 'field': 'Obs'},
            'List': {'w': 35},
            'Added': {'w': 65, 'sort': {'DateFormat': '%d %b'}},
            'Notes': {'w': 100, 'input': True},
            'Other': {'w': 1, 'align': 'left'},
        }

        # time control widget
        ctrl = BoxLayout(orientation='horizontal', size_hint=(1, 1))

        # observing list buttons
        ctrl.add_widget(
            CButton(
                text='add to list', width=dp(100), on_press=self.add_to_observing_list
            )
        )
        ctrl.add_widget(
            CButton(
                text='remove from list',
                width=dp(130),
                on_press=self.remove_from_observing_list,
            )
        )

        ctrl.add_widget(Label(size_hint_x=None, width=dp(40)))

        self.sun_time = TableLabel(text='', markup=True, size_hint_x=None, width=dp(80))

        ctrl.add_widget(self.sun_time)
        self.slider = Slider(
            orientation='horizontal',
            min=0,
            max=24,
            value=0,
            size_hint_x=None,
            width=dp(250),
        )
        self.slider.bind(value=self.time_changed)
        self.time_field = TableLabel(
            text=datetime.now().strftime('%d %B %H:%M'), size_hint_x=None, width=dp(160)
        )
        ctrl.add_widget(self.slider)
        ctrl.add_widget(self.time_field)
        self.time_changed(self.time_field, 0)

        Logger.debug(
            'ObservingList: built in {:.0f} ms'.format(1000 * (time.time() - t0))
        )

        return Table(
            size=Window.size,
            data=self.objects,
            name='DSOs',
            cols=cols,
            update_on_show=False,
            controls=ctrl,
            on_hide_method=self.app.table_hiding,
            initial_sort_column='RA',
        )

    def time_changed(self, widgy, value):
        # user moves time slider
        if hasattr(self, 'update_event'):
            self.update_event.cancel()
        t0 = datetime.now() + timedelta(seconds=3600 * value)
        self.time_field.text = t0.strftime('%d %B %H:%M')

        sun_alt = sun_altitude(t0, self.latitude, self.longitude)
        self.sun_time.text = 'sun: {:3.0f}\u00b0'.format(sun_alt)
        try:
            self.compute_altaz(current_hours_offset=self.slider.value)
        except Exception as e:
            Logger.error('ObservingList: problem computing altaz {:}'.format(e))

        if hasattr(self, 'table'):
            # update display since this is fast, but update table (reapply filters) when slider stops
            self.table.update_display()
            self.update_event = Clock.schedule_once(self.table.update, 0.5)

    def show(self, *args):
        '''Called from menu to browse DSOs; open on first use'''
        if not hasattr(self, 'table'):
            self.table = self.build()
        self.app.showing = 'observing list'

        # redraw on demand when required
        if self.table not in self.app.gui.children:
            self.app.gui.add_widget(self.table, index=0)

        self.table.show()
        #  update time
        self.time_changed(self.time_field, 0)
        self.update_status()

    def new_from_list(self, row, *args):
        # User selects a row in the observing list table
        name = row.fields['Name'].text + '/' + row.fields['OT'].text
        self.table.hide()
        Component.get('DSO').on_new_object(self.lookup_name(name))

    def update_status(self):
        if hasattr(self, 'objects') and len(self.objects) > 0:
            self.info(
                '{:} out of {:} dsos'.format(
                    len([1 for v in self.objects.values() if v['List'] == 'Y']),
                    len(self.objects),
                )
            )

    def add_to_observing_list(self, *args):
        dn = datetime.now().strftime('%d %b')
        for s in self.table.selected:
            self.objects[s]['Added'] = dn
            self.objects[s]['List'] = 'Y'
        Logger.info('ObservingList: added {:} objects'.format(len(self.table.selected)))
        self.update_list()

    def remove_from_observing_list(self, *args):
        for s in self.table.selected:
            self.objects[s]['Added'] = ''
            self.objects[s]['List'] = 'N'
        Logger.info(
            'ObservingList: removed {:} objects'.format(len(self.table.selected))
        )
        self.update_list()

    def new_observation(self):
        OT = Component.get('Metadata').get('OT', None)
        Name = Component.get('Metadata').get('Name', None)
        #  update observed count if DSO is known
        if Name is not None and OT is not None:
            name = '{:}/{:}'.format(Name, OT)
            if name in self.objects:
                self.objects[name]['Obs'] += 1
            if hasattr(self, 'table'):
                self.table.update()

    def update_list(self):
        self.save_observing_list()
        self.table.update()
        self.table.deselect_all()
        self.update_status()

    def lookup_name(self, nm, *args):
        if not hasattr(self, 'objects'):
            self.load()

        if nm in self.objects:
            od = self.objects[nm]

        else:  #  search thru names
            d = [dd for dd in self.objects.values() if dd['Name'].lower() == nm.lower()]
            if len(d) == 0:
                return None
            else:
                od = d[0]
        od['RA'] = RA(od['RA'])
        od['Dec'] = Dec(od['Dec'])
        return od

    # called once we are sure we have an object
    def lookup_details(self, nm):
        return self.objects.get(nm, {})

    def lookup(self, nm, max_matches=11):
        if not hasattr(self, 'objects'):
            self.load()
        nm = nm.lower()
        matches = [n for n in self.objects.keys() if n.lower().startswith(nm)]
        # reorder so that any exact matches ie starting with nm and slash, are prioritised
        priors = [n for n in matches if n.lower().startswith(nm + '/')]
        nonpriors = [n for n in matches if not n.lower().startswith(nm + '/')]
        all_matches = priors + nonpriors
        return all_matches[:max_matches]




''' Maintains and applies bad pixel map that persists between sessions
    new in V0.3: creates a new BPM for each sensor shape encountered to avoid applying
    BPM to the wrong sensor and to enable multiple sensors to be used 
'''

import os
import numpy as np
from scipy.ndimage import convolve

from kivy.app import App
from kivy.properties import ConfigParserProperty
from kivy.logger import Logger

from jocular.component import Component

class BadPixelMap(Component):

    apply_BPM = ConfigParserProperty(1, 'BadPixelMap', 'apply_BPM', 'app', val_type=int)
    bpm_frames = ConfigParserProperty(3, 'BadPixelMap', 'bpm_frames', 'app', val_type=int)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.bpm = None

    def load_BPM(self, shape):
        # load BPM for this shape, creating if necessary

        self.bplist = []  # acts as a circular buffer of BPs from recent frames
        self.frame_count = 0
        self.bpm = None
        self.bpm_shape = shape
        self.bpm_name = 'BPM_{:}x{:}.npy'.format(shape[0], shape[1])
        path = os.path.join(self.app.get_path('calibration'), self.bpm_name)
        if os.path.exists(path):
            bpm = np.load(path)
            self.update_bpm({(x, y) for x, y in zip(bpm[0], bpm[1])})
            self.info('loaded')
            Logger.info('BPM: loaded map with {:} members'.format(len(bpm[0])))
        else:
            self.info('new map')
            Logger.info('BPM: new map')

    def on_close(self):
        self.save_BPM()

    def process_bpm(self, sub):

        if not self.apply_BPM:
            return

        # we don't have one yet, so create one of correct shape
        if self.bpm is None:
            self.load_BPM(sub.shape)

        # we have one, but it is the wrong shape (ie different sensor) so save & load
        elif self.bpm_shape != sub.shape:
            self.save_BPM()
            self.load_BPM(sub.shape)

        im = sub.get_image()
        badpix = self.find_hot_pixels(im)

        # only update map if it is a light sub
        if sub.sub_type == 'light':
            self.update_bpm(badpix)
            if self.bpm is not None:
                self.info('{:} pix, {:} in map'.format(len(badpix), len(self.bpm)))
        self.do_bpm(im, self.bpm)


    def find_hot_pixels(self, im):
        '''Return hot pixel candidates.
        We look for non-edge pixels whose intensity is significantly greater
        than their neighbours. Approach is conservative in order to
        find all hot pixels, since (a) we remove any imposters by comparing
        across most recent 'bpm_frames' when building the hot pixel map; and (b) in the
        EAA use case the only adverse effect is to replace a few non-hot pixels
        by their median. Returns a set of (row, col) coordinates
        ''' 
                
        # set min to zero
        im_norm = im - np.min(im)

        # divide im by local sum in 3x3 region
        im2 = im_norm / convolve(im_norm, np.ones((3, 3)), mode='constant')

        # Define hot pix as more than 5 SD from mean
        hp_cands = (im2 > np.mean(im2) + 5*np.std(im2))

        # set boundaries to zero
        hp_cands[0, :] = 0
        hp_cands[-1, :] = 0
        hp_cands[:, 0] = 0
        hp_cands[:, -1] = 0

        # coordinates of hot pixels as 2 x N array
        hps = np.where(hp_cands>0)

        # return hot pixel coordinates as a set of row-col pairs
        return {(r, c) for r, c in zip(hps[0], hps[1])}

    def do_bpm(self, im, bpm=None):
        # Replace each pixel in bad pixel map by median of neighbours
        if bpm is None:
            bpm = self.bpm
        if bpm is not None:
            for r, c in bpm:
                im[r, c] = np.median(im[r-1:r+2, c-1:c+2].ravel())
        return im
 
    def compute_bpm(self):
        # Use intersection of bad pixels from previous N frames to compute bad pixel map
        if self.bplist:
            self.bpm = self.bplist[0]
            for bpl in self.bplist[1:]:
                self.bpm = self.bpm.intersection(bpl)

    def update_bpm(self, bpm):
        # Add new BPM set to BPM, recomputing BPM
        if len(self.bplist) == self.bpm_frames:
            self.bplist[self.frame_count % self.bpm_frames] = bpm
        else:
            self.bplist.append(bpm)
        self.frame_count += 1
        self.compute_bpm()

    def save_BPM(self):
        # Save as npy file
        if self.bpm is None:
            return
        bpm = np.array([[x for x, y in self.bpm], [y for x, y in self.bpm]])
        path = os.path.join(self.app.get_path('calibration'), self.bpm_name)
        np.save(path, bpm)
        Logger.info('BPM: saved map with {:} members'.format(len(bpm[0])))



''' Manages all individual components, including imports/instantiation,
    in a uniform way. A bit 'meta' but some advantages of doing it this
    way.
'''

import time
import importlib
import math
from collections import OrderedDict
from kivy.app import App
from kivy.logger import Logger
from kivy.event import EventDispatcher
from kivy.properties import StringProperty, NumericProperty, BooleanProperty
from jocular.widgets import jicon

def remove_nulls_from_dict(d):
    return {k: v for k, v in d.items() \
        if not((v is None) or \
            (isinstance(v, str) and v.strip() == '') or  \
            (isinstance(v, float) and math.isnan(v)) or  \
            (isinstance(v, dict) and len(v) == 0) or \
            (isinstance(v, list) and len(v) == 0))}

class Component(EventDispatcher):

    components = OrderedDict()
    infoline = StringProperty('')   # ensures each component can signal its status for technical panel
    errors = NumericProperty(0)     # use these at some point
    warnings = NumericProperty(0)
    changed = BooleanProperty(True)

    @classmethod
    def get(cls, name):
        # Returns component whose class is identified by name; loads if not found

        if name in cls.components:
            return cls.components[name]
        try:
            t0 = time.time()
            module = importlib.import_module('jocular.' + name.lower())
            class_ = getattr(module, name)
            import_time = time.time() - t0
            t1 = time.time()
            inst = class_()
            init_time = time.time() - t1
            cls.register(inst)
            Logger.info('Component: imported {:} in {:.0f}ms, init in {:.0f}ms'.format(name, 
                1000 * import_time, 1000 * init_time))
            return inst
        except Exception as e:
            Logger.error('Component: cannot import module {:} ({:})'.format(name, e))
            return
 
    @classmethod
    def is_loaded(cls, name):
        return name in cls.components

    @classmethod
    def register(cls, obj):
        name = obj.__class__.__name__
        if name not in cls.components:
            cls.components[name] = obj
            App.get_running_app().gui.initialise_component(name)
        else:
            Logger.error('Component: name clash for {:}'.format(name))
 
    @classmethod
    def initialise_new_object(cls):
        for c in cls.components.keys():
            cls.components[c].changed = False
            cls.components[c].on_new_object()

    @classmethod
    def save_object(cls):
        for c in cls.components.keys():
            cls.components[c].on_save_object()

    @classmethod
    def close(cls):
        Logger.info('Component: closedown')
        for c in cls.components.keys():
            cls.components[c].on_close()

    @classmethod
    def bind_status(cls):
        for name, inst in cls.components.items():
            cls.get('Status').bind_status(name, inst)
            inst.info('loaded')

    @classmethod
    def check_change(cls):
        for c in cls.components.keys():
            if cls.components[c].changed:
                App.get_running_app().gui.something_has_changed = True
                return
        App.get_running_app().gui.something_has_changed = False

    def redraw(self, *args): pass
    def on_new_object(self, *args): pass
    def on_save_object(self, *args): pass
    def on_close(self, *args): pass

    def on_changed(self, *args):
        Component.check_change()

    def info(self, message=None, prefix=None, typ='normal'):
        i = '' if typ == 'normal' else jicon(typ)
        self.infoline = '{:}: [b]{:}[/b] {:}'.format(message, self.__class__.__name__, i)

    def warn(self, message):
        self.warnings += 1
        self.info(message, typ='warn')
        Logger.debug('{:}: {:}'.format(self.__class__.__name__, message))

    def error(self, message):
        self.errors += 1
        self.info(message, typ='error')
        Logger.exception('{:}: {:}'.format(self.__class__.__name__, message))




''' Class representing sorteable, selectable table view associated with a json model.
'''

import math
import os
import time
import operator
from functools import partial
from datetime import datetime
import numpy as np

from kivy.app import App
from kivy.logger import Logger
from kivy.core.window import Window
from kivy.uix.floatlayout import FloatLayout
from kivy.uix.boxlayout import BoxLayout
from kivy.uix.checkbox import CheckBox
from kivy.uix.button import Button
from kivy.uix.textinput import TextInput
from kivy.uix.label import Label
from kivy.lang import Builder
from kivy.properties import NumericProperty, StringProperty, BooleanProperty
from kivy.metrics import dp
from kivy.clock import Clock

Builder.load_string('''

<TableFormat>:
    font_size: '15sp'
    margin: dp(5)
    padding: dp(5), dp(5)
    text_size: self.size
    color: .5, .5, .5, 1
    valign: 'middle'

<HeaderButton>:
    on_press: root.sort_column()
    browser: None
    halign: 'center'
    background_color: .08, .08, .08, 1
    color: app.highlight_color

# row label
<TableLabel>:
    background_normal: ''
    height: dp(24)

# name column button
<TableButton>:
    column: ''
    background_normal: ''
    background_disabled_normal: ''
    background_color: 0, 0, 0, 0

<TableInput>:
    background_normal: ''
    height: dp(24)
    background_color: app.background_color
    foreground_color: app.lowlight_color
    disabled_foreground_color: app.lowlight_color
    multiline: False
    halign: 'right'
    hint_text_color: app.hint_color
    font_size: '16sp'
    padding: dp(0), dp(2), dp(0), dp(2)

# control button
<CButton>:
    size_hint: None, 1
    width: dp(180)
    text_size: self.size
    browser: None
    halign: 'center'
    #background_normal: ''
    color: app.highlight_color
    background_color: .06, .06, .06, 0

<SearchBarTextInput@TextInput>:
    background_color: .1, .1, .1, 1
    foreground_color: .5, .5, .5, 1
    multiline: False
    font_size: '16sp'

<SearchBar>:
    canvas.after:
        Color:
            rgba: app.line_color
        Line:
            points: self.x, self.y + dp(2), self.x + self.width, self.y + dp(2)
            width: 1
    size_hint: 1, None
    height: dp(32)
    font_size: '20sp'
    orientation: 'horizontal'

<TableHeader>:
    canvas.after:
        Color:
            rgba: app.line_color
        Line:
            points: self.x, self.y + dp(2), self.x + self.width, self.y + dp(2)
            width: dp(1)
    size_hint: 1, None
    height: dp(32)
    font_size: '20sp'
    orientation: 'horizontal'

<TableRow>:
    size_hint: 1, None
    height: dp(28)
    orientation: 'horizontal'

<Table>:
    size: dp(1100), dp(600)
    contents: _contents
    myheader: _header
    footer: _footer
    spacer: _spacer
    footer_actions: _footer_actions

    BoxLayout:
        orientation: 'vertical'
        pos: root.pos
        canvas.before:
            Color:
                rgba: .08, .08, .08, 1 
            Rectangle:
                pos: self.pos
                size: self.size

        BoxLayout:
            id: _header
            size_hint: 1, None
            height: dp(64)
            orientation: 'vertical'

        BoxLayout:
            id: _allcontents
            orientation: 'horizontal'
            size_hint: 1, 1                

            BoxLayout:
                id: _contents
                size_hint: None, 1
                width: root.width - dp(30)
                orientation: 'vertical'

            Slider:
                size_hint: None, 1
                cursor_size: [dp(25), dp(25)]
                width: dp(30)
                disabled: root.n_matching < root.n_rows
                min: 0
                max: max(0, root.n_matching - root.n_rows)
                step: 1
                value: round(max(0, root.n_matching - root.n_rows) - root.current_row)
                on_value: root.current_row = round(max(0, root.n_matching - root.n_rows) - self.value)
                orientation: 'vertical'
                padding: dp(16)

        BoxLayout:
            id: _spacer
            size_hint: 1, None
            canvas.after:
                Color:
                    rgba: app.line_color
                Line:
                    points: self.x, self.y + self.height, self.x + self.width, self.y + self.height
                    width: 1
            size_hint: 1, None
            height: dp(4)
            orientation: 'horizontal'

        BoxLayout:
            id: _footer
            size_hint: 1, None
            height: dp(32)
            orientation: 'horizontal'

            CButton:
                width: dp(80)
                text: 'select all'
                on_press: root.select_all()

            CButton:
                text: 'deselect all'
                width: dp(95)
                on_press: root.deselect_all()

            TableLabel:
                size_hint: None, 1
                width: dp(120)
                text: '{:d} of {:d}'.format(root.n_matching, root.n_rows_total)

            BoxLayout:
                id: _footer_actions
                height: dp(32)
                orientation: 'horizontal'
                
            CButton:
                width: dp(70)
                text: 'export'
                on_press: root.export(self)

            CButton:
                width: dp(50)
                text: 'close'
                halign: 'right'
                on_press: root.hide()

''')

class TableFormat:
    pass

# header column button (controls sorting)
class HeaderButton(Button, TableFormat):

    def __init__(self, table=None, column_name=None, column_type=None, sort_type=None, **kwargs):
        super().__init__(**kwargs)
        self.column_name = column_name
        self.table = table
        self.column_type = column_type
        self.sort_type = sort_type

    def sort_column(self):
        self.table.current_sort = {
            'name': self.column_name,
            'type': self.column_type,
            'sort_type': self.sort_type
        }
        self.table.sort_column()

# ordinary label
class TableLabel(Label, TableFormat):
    pass

# control button at lower of screen
class CButton(Button, TableFormat):
    pass

# ordinary button
class TableButton(Button, TableFormat):
    pass

# ordinary button
class TableInput(TextInput, TableFormat):
    pass

# table header line i.e. click-to-sort column headers -----------------------

class TableHeader(BoxLayout):

    def __init__(self, table, **kwargs):
        super().__init__(**kwargs)
        self.table = table

        # selection column
        self.add_widget(HeaderButton(size_hint_x=None, width=dp(20), halign='center', 
            table=table, text=' ', column_name='Selected', column_type=str))

        for i, (colname, colprops) in enumerate(table.cols.items()):

            width = colprops.get('w', 100)
            self.add_widget(HeaderButton(
                text=colname,
                size_hint_x=width if width < 2 else None,
                width=dp(width) if width >= 2 else width,
                halign=colprops.get('align', 'center'),
                column_name=colprops.get('field', colname),
                column_type=colprops.get('type', str),
                sort_type=colprops.get('sort', 'normal'),
                table=table
                ))

    def resize(self, *args):
        pass

# a single row of the table ----------------------

class TableRow(BoxLayout):

    def __init__(self, table, **kwargs):

        self.table = table
        super().__init__(**kwargs)

        self.updating = True  # flag to prevent updating textinput fields during redrawing

        # construct selector column
        w = CheckBox(size_hint_x=None, width=dp(20))
        w.bind(active=self.select_row)
        self.add_widget(w)

        self.fields = {'Selected': w}

        for colname, colprops in self.table.cols.items():

            width = colprops.get('w', 100)

            # input field; any text gets saved to self.data[key][colname]
            if 'input' in colprops:
                w = TableInput(size_hint_x=width if width < 2 else None,
                    width=dp(width) if width >= 2 else width)
                w.bind(text=partial(self.input_changed, colprops.get('field', colname)))

            else:              
                if 'action' in colprops:
                    wtype = TableButton
                else:
                    wtype = TableLabel

               # width = colprops.get('w', 100)
                w = wtype(
                    text='',
                    size_hint_x=width if width < 2 else None,
                    width=dp(width) if width >= 2 else width,
                    halign=colprops.get('align', 'center'),
                    shorten=True,
                    shorten_from='right')

                if 'action' in colprops:
                    w.bind(on_press=self.row_clicked)  # add partial with column name

            self.add_widget(w)
            self.key = None

            self.fields[colprops.get('field', colname)] = w 

        self.updating = False


    def row_clicked(self, *args):
        # only supports one at present <-- need to fix this using partial
        for colname, colprops in self.table.cols.items():
            if 'action' in colprops:
                colprops['action'](self)
                return

    # prevent click thru to beneath
    def on_touch_down(self, touch):
        if self.table.hidden:
            return
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled

    def select_row(self, *args):
        if self.fields['Selected'].active:
            self.table.selected.add(self.key)
        elif self.key in self.table.selected:
            self.table.selected.remove(self.key)

    def input_changed(self, colname, widgy, value):
        if not self.updating:
            if self.key is not None:
                if self.key in self.table.data:
                    self.table.data[self.key][colname] = value
 
    def update_row(self, row, key=None, empty=False):

        self.updating = True # don't allow input_changed to fire while redrawing
        if not empty:
            self.key = key
            self.fields['Selected'].active = self.key in self.table.selected
            self.fields['Selected'].disabled = False
        else:
            self.fields['Selected'].active = False
            self.fields['Selected'].disabled = True

        for colname, colprops in self.table.cols.items():
            field = colprops.get('field', colname)
            display_fn = colprops.get('display_fn', str)
            self.fields[field].text = ''
            if not empty and (field in row):
                val = row[field]
                vtype = type(val)
                if vtype == str:
                    self.fields[field].text = display_fn(val)
                elif not np.isnan(val):
                    self.fields[field].text = display_fn(val)
        self.updating = False

#-------------------------------------------------------------------------------------------
# table search line i.e. column filters

class SearchBar(BoxLayout):

    def __init__(self, table, **kwargs):

        self.table = table
        super().__init__(**kwargs)

        self.sorted_column = 'Name'
        self.reverse_sort = {'Name': False}

        self.filters = []

        w = SearchFilter(text='', size_hint_x=None, width=dp(20), 
            table=table, column='Selected', column_type=str)
        self.add_widget(w)
        # self.filters.append(w)

        for colname, colprops in self.table.cols.items():
            width = colprops.get('w', 100)
            w = SearchFilter(
                text='',
                size_hint_x=dp(width) if width < 2 else None,
                width=dp(width) if width >= 2 else 1,
                column=colprops.get('field', colname),
                table=table,
                val_fn=colprops.get('val_fn', None),
                column_type=colprops.get('type', str)
                )
            self.add_widget(w)
            self.filters.append(w)

        self.add_widget(CButton(text='clear', width=dp(50), on_press=self.clear_search))

    def clear_search(self, *args):
        for f in self.filters:
            f.text = ''
        self.table.search_results = list(self.table.data.keys())
        self.table.on_search_results()

    def resize(self, *args):
        pass

    def combine_results(self):
        t0 = time.time()
        # these two lines ~ 10ms
        l = [s.filtered for s in self.filters]
        self.table.search_results = list(set.intersection(*l)) # causes update
        # this takes ~ 50ms
        self.table.on_search_results()

    def reapply_filters(self):
        # called whenever the underlying table data has changed
        for f in self.filters:
            f.has_had_text = False
            f.on_text(f, f.text)
        self.combine_results()

#-------------------------------------------------------------------------------------------
# an individual column filter

# support methods to extract args/operators for filtering operations
def parg(s):
    # parse single arg
    s = s.strip()
    try:
        if s[0] not in '=!<>':
            s = '=' + s
        rest = s[1:].strip()
        if rest:
            return {s[0]: rest}
        else:
            return {}
    except:
        return {}
    
def parse_args(s):
    # parse normal or comma-sep arg
    try:
        s = s.strip()
        if ',' not in s:
            return parg(s)
        p1, _, p2 = s.strip().partition(',')
        p1 = p1.strip()
        p2 = p2.strip()
        d = {}
        if p1 and (p1[0] not in '<>'):
            d['>'] = p1
        elif p1:
            d[p1[0]] = p1[1:].strip()
        if p2 and (p2[0] not in '<>'):
            d['<'] = p2
        elif p2:
            d[p2[0]] = p2[1:].strip()
        return d
    except:
        return {}

class SearchBarTextInput(TextInput):
    pass

class SearchFilter(SearchBarTextInput):

    def __init__(self, table=None, column=None, column_type=None, val_fn=None, **kwargs):

        self.table = table
        self.column = column
        self.column_type = column_type
        self.val_fn = val_fn
        self.has_had_text = False   # see comment in on_text for the why
        super().__init__(**kwargs)

        # current set that matches filter for this column
        self.filtered = set({})


    def on_text(self, instance, value):
        ''' Applies search filter when user types. We use a flag to indicate if 
        the search field has ever contained text. If so, when it becomes empty 
        is is necessary to update the results. Otherwise, we don't update on empty 
        to avoid multiple identical updates.
        '''

        ''' Timing-wise, the search is pretty fast actually about 30ms for 40k rows; takes longer to then
            intersect the results and redraw; could be worth optimising for large catalogues, but not yet
        '''


        value = value.strip()

        if not value:
            self.filtered = set(self.table.data.keys())
            if self.has_had_text:
                self.table.searchbar.combine_results()
            return

        try:

            objs = self.table.data
            col = self.column
            col_type = self.column_type

            if len(value) > 0:

                self.has_had_text = True

                # select column where it exists
                objs = {k: v[col] for k, v in objs.items() if col in v}

                if self.column_type == str:

                    objs = {k: v.lower() for k, v in objs.items()}
                    value = value.lower()

                    # to do: had ! or != for not equal

                    # use wildcard syntax, and = for exact match
                    if value[0] == '*':
                        if len(value) > 1:
                            if value[-1] == '*':
                                objs = {k: v for k, v in objs.items() if value[1:-1] in v}
                            else:
                                objs = {k: v for k, v in objs.items() if v.endswith(value[1:])}
                    elif value[0] == '=':
                        objs = {k: v for k, v in objs.items() if v == value[1:]}
                    elif value[0] == '!':
                        objs = {k: v for k, v in objs.items() if v != value[1:]}
                    else:
                        if value[-1] == '*':
                            value = value[:-1]
                        objs = {k: v for k, v in objs.items() if v.startswith(value)}


                elif self.column_type in [float, int]:

                    objs = {k: col_type(v) for k, v in objs.items()}
                    col = self.column

                    ''' allow expressions of the form =x, !x or
                        intervals using either x,y or >x, <y
                        (or just >x for example). If the interval
                        is 'normal' e.g. 3 < x < 8 then choose x
                        in range [3, 8]; if interval is exclusive
                        e.g. <3, >10, treat as an 'or' and return
                        objects that satisfy either constraint. This
                        is useful for situations like hours spanning
                        midnight
                    '''

                    # can use x,y >x, <y =x, !x (not equal)
                    ops = parse_args(value)
                    for k, v in ops.items():
                        #ops[k] = self.column_type(v)
                        ops[k] = float(v)

                    # check if it is and or an (implicit) or
                    if ('<' in ops) and ('>' in ops) and (ops['<'] < ops['>']):
                        # it is an or
                        comp_value = ops['<']
                        if self.val_fn:
                            comp_value = self.val_fn(comp_value)
                        objs1 = {k: v for k, v in objs.items() if v <= comp_value}
                        comp_value = ops['>']
                        if self.val_fn:
                            comp_value = self.val_fn(comp_value)
                        objs2 = {k: v for k, v in objs.items() if v >= comp_value}
                        objs = {**objs1, **objs2}

                    else:
                        # ops is a dict with keys in 
                        #   =, !, >, <
                        for op, comp_value in ops.items():
                            if self.val_fn:
                                comp_value = self.val_fn(comp_value)
                            if op == '=':
                                objs = {k: v for k, v in objs.items() if v == comp_value}
                            elif op == '!':
                                objs = {k: v for k, v in objs.items() if v != comp_value}
                            elif op == '>':
                                objs = {k: v for k, v in objs.items() if v >= comp_value}
                            elif op == '<':
                                objs = {k: v for k, v in objs.items() if v <= comp_value}

            self.filtered = set(objs.keys())

            # update search results by intersecting with current search results from other columns
            # also does the redraw so quite expensive
            self.table.searchbar.combine_results()

        except Exception:
            pass

#-------------------------------------------------------------------------------------------
# Main Table class

class Table(FloatLayout):

    n_rows = NumericProperty(0)      # num visible rows
    n_matching = NumericProperty(0)  # number matching search/filter criteria
    current_row = NumericProperty(0)
    n_rows_total = NumericProperty(0)
    name = StringProperty('')
    description = StringProperty('')
    hidden = BooleanProperty(True)

    def __init__(self, size=None, data=None, cols=None, name='', 
        actions=None, controls=None,
        description='',
        update_on_show=True, on_show_method=None, on_hide_method=None, 
        initial_sort_column=None, initial_sort_direction=None, **kwargs):


        t0 = time.time()
        self.initialising = True  # to prevent redraw
        super().__init__(**kwargs)

        self.data = data
        self.n_rows_total = len(self.data)   # need to ensure this is kept updated
        self.cols = cols
        self.name = name
        self.description = description

        # user can either provide a set of button labels and method calls
        # via actions
        self.actions = actions

        # or (new in v0.3) a widget that is handled externally to table
        self.controls = controls

        self.update_on_show = update_on_show
        self.selected = set({})
        self.on_show_method = on_show_method
        self.on_hide_method = on_hide_method

        self.initial_sort_direction = initial_sort_direction
        if initial_sort_column is not None:
            if initial_sort_column in cols:
                vals = cols[initial_sort_column]
                self.current_sort = {
                    'name': initial_sort_column,
                    'type': vals.get('type', str),
                    'sort_type': vals.get('sort','')
                }

        self.filters = {}
        self.reverse_sort = {}
        self.search_results = list(self.data.keys())

        self.initialising = False  # now allow redraw
        self.size = Window.size
        self.redraw()
        # the redraw takes the time
        Logger.debug('Table: {:} built in {:.0f} ms'.format(name, 1000*(time.time() - t0)))


    def redraw(self, *args):

        if self.initialising:
            return

        t0 = time.time()

        self.redrawing = True

        if not hasattr(self, 'header'):
            # generate header/searchbar components
            self.header = TableHeader(self)
            self.myheader.add_widget(self.header)
            self.searchbar = SearchBar(self)
            self.myheader.add_widget(self.searchbar)
            self.empty_row = TableRow(self)
        else: # remove rows
            for r in self.rows:
                self.contents.remove_widget(r)

        # new in v0.5
        h = self.height - self.myheader.height - self.footer.height
        nrows = int(h / self.empty_row.height)
        spare = h - nrows * self.empty_row.height
        self.rows = []
        for i in range(0, nrows):
            r = TableRow(self)
            self.rows.append(r)
        self.n_rows = len(self.rows)

        for r in self.rows:
            self.contents.add_widget(r)
        self.spacer.height = spare

        # about 300 ms for DSO table
        self.footer_actions.clear_widgets()
        if self.actions:
            for k, v in self.actions.items():
                self.footer_actions.add_widget(CButton(text=k, on_press=v))

        elif self.controls:
            self.footer_actions.add_widget(self.controls)
 
        self.redrawing = False

        # about 100 ms for DSO table
        self.on_search_results()

        if self.initial_sort_direction is not None:
            self.sort_column(reverse=self.initial_sort_direction)


    def on_search_results(self, *args):

        if self.initialising: #  or self.redrawing:
            return

        self.current_row = 0
        self.n_matching = len(self.search_results)
        self.sort_column(reverse=False)
        self.update_display()

    def update(self, *args):
        # performs immediate update of table contents following a change
        # typically called on show and also from clients who make changes to data

        self.selected = {s for s in self.selected if s in self.data}
        self.n_rows_total = len(self.data)
        if hasattr(self, 'searchbar'):
            self.searchbar.reapply_filters()

    def update_display(self, *args): 
        # populate screen cells with current data

        last_row = min(self.n_matching, self.current_row + self.n_rows)
        results = self.search_results[self.current_row: last_row]

        for i, r in enumerate(results):
            # catch odd bug and log it
            if r in self.data:
                self.rows[i].update_row(self.data[r], key=r)
            else:
                Logger.debug('Table: cannot find {:} in table'.format(r))

        for j in range(last_row - self.current_row, self.n_rows):  # clear rest
            self.rows[j].update_row(self.empty_row, empty=True)

    def show(self, *args):

        # only allow resize when it is showing so as not to slow down general resizing...
        self.bind(size=self.redraw)

        # redraw if size has changed since last hide
        self.pos = 0, 0
        if hasattr(self, 'size_on_hide') and self.size_on_hide != self.size:
            self.redraw()

        if self.on_show_method:
            self.on_show_method()

        if self.update_on_show:
            self.update()
        else:
            self.update_display()

        self.hidden = False

    def hide(self, *args):

        self.unbind(size=self.redraw)
        if self.on_hide_method:
            self.on_hide_method()
        self.x = -10 * self.width
        self.size_on_hide = self.size.copy() # don't use size as it is a reference
        self.hidden = True

    def on_current_row(self, *args):
        self.update_display()

    def select_all(self, *args):
        self.selected = set(self.search_results)
        self.update_display()

    def deselect_all(self, *args):
        self.selected = set({})
        self.update_display()

    def select(self, to_select):
        self.selected = set(to_select)
        self.update_display()

    def export(self, but):
        def no_nl(s):
            return s.replace('\n', ' ')

        self.export_button = but
        self.orig_export_button_color = but.color
        but.color = [1, 0, 0, 1]
        cols = self.cols.keys()
        when = datetime.now().strftime('%d_%b_%y_%H_%M')

        path = os.path.join(App.get_running_app().get_path('exports'), 
            '{:}_{:}.csv'.format(self.name.replace(' ','_'), when))
        with open(path, 'w') as f:
            f.write(','.join(cols) + '\n')
            cols = [v.get('field', k) for k, v in self.cols.items()]
            for s in self.search_results:
                sd = self.data[s]
                f.write(','.join([no_nl(str(sd.get(c,''))) for c in cols]))
                f.write('\n')
        Clock.schedule_once(self.change_export_button, 1)

    def change_export_button(self, dt):
        self.export_button.color = self.orig_export_button_color

    def get_selected(self):
        return self.selected

    def on_touch_down(self, touch):
        if self.hidden:
            return
        handled = super().on_touch_down(touch)
        if self.collide_point(*touch.pos):
            return True
        return handled

    def sort_column(self, reverse=True):

        def defaultsplit(x, default):
            a = x.split()
            if len(a) == 0:
                return '', default
            elif len(a) == 1:
                return a[0], default
            elif a[1].isdigit():
                return a[0], float(a[1])
            elif a[0].isdigit():
                return float(a[0]), a[1]
            else:
                return x, default

        def null_value(v):
            return (isinstance(v, str) and v.strip() == '') or (isinstance(v, float) and math.isnan(v))

        if self.initialising or self.redrawing:
            return

        if not hasattr(self, 'current_sort'):
            return

        column_name = self.current_sort['name']
        column_type = self.current_sort['type']
        sort_type = self.current_sort['sort_type']

        # toggle sort direction
        if column_name in self.reverse_sort:
            if reverse:
                self.reverse_sort[column_name] = not self.reverse_sort[column_name]
        else:
            self.reverse_sort[column_name] = False

        # current filtered results
        current = np.array(self.search_results)

        # set up dict from position to value for current column, where column exists
        x = {i: self.data[k][column_name] for i, k in enumerate(current) if column_name in self.data[k]}

        # remove any null values that have crept in
        # x = {i: v for i, v in x.items() if v}
        x = {i: v for i, v in x.items() if not null_value(v)}

        if not x:
            return

        # store indices that won't be sorted as they don't have a value
        blank_indices = list(set(np.arange(len(self.search_results))) - set(x.keys()))

        # catch specialised sorts first
        if type(sort_type) == dict:
            if 'catalog' in sort_type: 
                # catalog sort sorts first and second columns e.g. NGC 1
                vals = [defaultsplit(str(v), 0) for v in x.values()]
                cat_sort_index = np.lexsort(([v[1] for v in vals], [v[0] for v in vals]))
                sort_index = np.array(list(x.keys()))[cat_sort_index]

            elif 'DateFormat' in sort_type:
                fmt = sort_type['DateFormat']
                x = {k: datetime.strptime(v, fmt) for k, v in x.items()}
                sort_index = [i for i, j in sorted(x.items(), key=operator.itemgetter(1))]

        elif column_type in [str, float, int]:
            x = {k: column_type(v) for k, v in x.items()}
            sort_index = [i for i, j in sorted(x.items(), key=operator.itemgetter(1))]
 
        else:
            sort_index = np.argsort(x.values())

        # reverse if necessary
        if self.reverse_sort[column_name]:
            sort_index = sort_index[::-1]

        # append blank indices
        self.search_results = list(current[sort_index]) + list(current[blank_indices])

        self.update_display()



''' Local plate-solver. 
'''


import numpy as np
from scipy.spatial.distance import cdist
from skimage.feature import blob_dog
from skimage.transform import estimate_transform
from concurrent.futures import ThreadPoolExecutor

from kivy.app import App
from kivy.logger import Logger
from kivy.properties import ConfigParserProperty

from jocular.component import Component
from jocular.RA_and_Dec import RA, Dec
from jocular.uranography import (
    to360,
    make_tile,
    Eq2Cart,
    Cart2Eq,
    spherical_distance,
    compute_centroids,
)


def select_stars_in_FOV(ras, decs, ra0, dec0, fov):
    #  filter ras, decs, mags to just return those in given region
    tile = make_tile(ra0, dec0, fov=fov)
    min_ra, max_ra = tile['min_ra'], tile['max_ra']
    min_dec, max_dec = tile['min_dec'], tile['max_dec']
    if min_ra < max_ra:
        locs = (ras >= min_ra) & (ras < max_ra) & (decs >= min_dec) & (decs < max_dec)
    else:
        locs = ((ras >= min_ra) | (ras < max_ra)) & (decs >= min_dec) & (decs < max_dec)
    return locs


def fastmatch(
    ref=None,
    im=None,
    mag_range=5,
    min_matches=8,
    proximity=0.05,
    match_arcsec=15,
    first_match=True,
):
    """find possible match of image points im in star database cloud db
    both db and im are ordered by magnitude; works by choosing a pair
    of image points, rotating and scaling so they sit at (0, 0) and (1, 0);
    then searching across pairs of image points at a similar separation, applying
    similar rotation/scaling, then counting matches; if first_match is true, return
    first match that contains more than min_matches stars, otherwise return
    the match that contains the most correspondences.
    """

    max_matches = 0
    best_matches = None
    n_im, n_ref = len(im), len(ref)
    for i1 in range(0, n_im):
        #  normalise image based on star index i1
        im1 = im - im[i1, :]
        for i2 in [i for i in range(i1 + 1, min(n_im, i1 + mag_range)) if i != i1]:
            # rotate and scale so i2 is at (1, 0)
            cos, sin = im1[i2, 0], -im1[i2, 1]
            d = cos ** 2 + sin ** 2
            im2 = np.dot(im1, np.array([[cos, -sin], [sin, cos]]).T) / d
            # match threshold takes into account scaling by d
            mt = (match_arcsec / (3600 * (d ** 0.5))) ** 2.0  # we match squared dist
            min_x, min_y = np.min(im2, axis=0) - mt
            max_x, max_y = np.max(im2, axis=0) + mt
            for r1 in range(0, n_ref):
                ref1 = ref - ref[r1, :]
                for r2 in [
                    r for r in range(r1 + 1, min(n_ref, r1 + mag_range)) if r != r1
                ]:
                    cos, sin = ref1[r2, 0], -ref1[r2, 1]
                    d2 = cos ** 2 + sin ** 2
                    if np.abs(d - d2) < proximity:
                        ref2 = np.dot(ref1, np.array([[cos, -sin], [sin, cos]]).T) / d2
                        mind_x, mind_y = np.min(ref2, axis=0)
                        maxd_x, maxd_y = np.max(ref2, axis=0)
                        #  don't check if anything outside range
                        if (
                            max_x < maxd_x
                            and max_y < maxd_y
                            and min_x > mind_x
                            and min_y > mind_y
                        ):
                            dists = cdist(im2, ref2, metric='sqeuclidean')
                            matches = [
                                (i, j)
                                for i, j in enumerate(np.argmin(dists, axis=1))
                                if dists[i, j] < mt
                            ]
                            n_matches = len(matches)
                            if n_matches >= max(min_matches, max_matches):
                                if first_match:
                                    return matches
                                max_matches = n_matches
                                best_matches = matches
                                # return matches
    return best_matches


class PlateSolver(Component):

    focal_length = ConfigParserProperty(
        800.0, 'PlateSolver', 'focal_length', 'app', val_type=float
    )
    pixel_height = ConfigParserProperty(
        8.4, 'PlateSolver', 'pixel_height', 'app', val_type=float
    )
    binning = ConfigParserProperty(1, 'PlateSolver', 'binning', 'app', val_type=int)
    n_stars_in_image = ConfigParserProperty(
        30, 'PlateSolver', 'n_stars_in_image', 'app', val_type=int
    )
    min_matches = ConfigParserProperty(
        8, 'PlateSolver', 'min_matches', 'app', val_type=int
    )
    match_arcsec = ConfigParserProperty(
        15, 'PlateSolver', 'match_arcsec', 'app', val_type=float
    )
    mag_range = ConfigParserProperty(5, 'PlateSolver', 'mag_range', 'app', val_type=int)
    first_match = ConfigParserProperty(
        1, 'PlateSolver', 'first_match', 'app', val_type=int
    )

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.app = App.get_running_app()
        self.cart2pix = None

    def on_new_object(self):
        self.cart2pix = None
        self.can_solve = Component.get('Catalogues').has_platesolving_db()
        self.info('ready' if self.can_solve else 'no database')

    def match(self):
        # runs in separate thread

        try:
            self.h, self.w = self.im.shape
            # degrees per pixel
            dpp = self.binning * (self.pixel_height * 1e-6 / self.focal_length) * (206265 / 3.6)
            fov = self.h * dpp

            #  extract image stars
            star_thresh = 0.001
            x, y, flux = compute_centroids(
                self.im,
                blob_dog(
                    self.im, min_sigma=3, max_sigma=5, threshold=star_thresh, overlap=0
                )[:, [1, 0]],
            )

            #  convert flux to relative magnitude
            mags = -2.5 * np.log10(flux / np.max(flux))

            # select N-brightest
            inds = np.argsort(mags)[: self.n_stars_in_image]
            x_im, y_im = x[inds], y[inds]

            if len(x_im) < self.min_matches:
                self.warn('too few stars ({:})'.format(len(x_im)))
                return

            # get reference stars for search field (NB larger fov is worse)
            ras, decs, mags = Component.get('Catalogues').get_platesolving_stars(
                make_tile(self.ra0, self.dec0, fov=2 * fov)
            )

            #  sort by magnitude and limit to N-brightest (should depend on area re stars extracted in central zone)
            inds = np.argsort(mags)[:80]

            #  convert to cartesian coords
            ras, decs = Eq2Cart(ras[inds], decs[inds], self.ra0, self.dec0)

            # do fastmatch with ref and im stars in degrees
            matches = fastmatch(
                ref=np.degrees(np.vstack([ras, decs]).T),
                im=dpp * np.vstack([x_im, y_im]).T,
                match_arcsec=self.match_arcsec,
                min_matches=self.min_matches,
                mag_range=self.mag_range,
                proximity=0.05,
                first_match=self.first_match,
            )

            # check if we have a result
            if matches is None:
                self.warn('failed to match')
                Logger.warn('Platesolver: no match, {:} im stars, {:} ref stars'.format(
                    len(x_im), len(ras)))
                return

            self.info('matched {:}'.format(len(matches)))

            # use result to find transform from ref in cartesian to im  in pixels
            src = np.vstack([ras, decs]).T[[j for (i, j) in matches], :]
            dst = np.vstack([x_im, y_im]).T[[i for (i, j) in matches], :]

            # consider doing RANSAC or alternative transform here
            # self.cart2pix = estimate_transform('similarity', src, dst)
            self.cart2pix = estimate_transform('affine', src, dst)

            # find centre of frame in RA/Dec
            self.tile_ra0, self.tile_dec0 = self.pixels_to_ra_dec(
                self.h // 2, self.w // 2
            )

            # estimate FOV
            self.FOV_h = spherical_distance(
                *self.pixels_to_ra_dec(0, 0), *self.pixels_to_ra_dec(self.h, 0)
            )
            self.FOV_w = spherical_distance(
                *self.pixels_to_ra_dec(0, 0), *self.pixels_to_ra_dec(0, self.w)
            )

            #  and compute where North is
            xx, yy = self.ra_dec_to_pixels(
                np.array([self.ra0, self.ra0]),
                np.array([self.dec0 - 0.5, self.dec0 + 0.5]),
            )
            self.north = to360(
                90 - np.degrees(np.arctan2(yy[1] - yy[0], xx[1] - xx[0]))
            )

            Logger.info(
                'PlateSolver: {:5.3f} x {:5.3f}, {:.0f}\u00b0, RA: {:} Dec: {:}'.format(
                    self.FOV_w,
                    self.FOV_h,
                    self.north,
                    str(RA(self.tile_ra0)),
                    str(Dec(self.tile_dec0)),
                )
            )

        except Exception as e:
            Logger.warn('PlateSolver: error in match')

    def solve(self, *args):
        # called when user presses loc icon

        self.cart2pix = None

        if not self.can_solve:
            return

        self.ra0, self.dec0 = Component.get('DSO').current_object_coordinates()
        if self.ra0 is None:
            self.warn('unknown DSO')
            return

        # get current image (sub or stack)
        self.im = Component.get('Stacker').get_image_for_platesolving()
        if self.im is None:
            self.warn('no current image')
            return

        pool = ThreadPoolExecutor(3)
        future = pool.submit(self.match)
        future.add_done_callback(self.annotate)

    def annotate(self, *args):
        if self.cart2pix is None:
            self.warn('failed')
        else:
            Component.get('Annotator').annotate()

    def ra_dec_to_pixels(self, ra, dec):
        # generate pixel coordinates corresponding to ra and dec
        # referenced to centre of closest matching tile
        if self.cart2pix is None:
            return
        if type(ra) != list and type(ra) != np.ndarray:
            ra = np.array([ra])
            dec = np.array([dec])

        x, y = Eq2Cart(ra, dec, self.ra0, self.dec0)
        d = self.cart2pix(np.array([x, y]).T)
        return d[:, 0], d[:, 1]

    def pixels_to_ra_dec(self, x, y):
        # generate pixel coordinates corresponding to ra and dec
        # referenced to centre of closest matching tile
        if self.cart2pix is None:
            return
        if type(x) != list and type(x) != np.ndarray:
            x = np.array([x])
            y = np.array([y])

        d = self.cart2pix.inverse(np.array([x, y]).T)
        x, y = Cart2Eq(d[:, 0], d[:, 1], self.ra0, self.dec0)
        if len(x) == 1:
            return x[0], y[0]
        return x, y

    # def get_stars(self, tile):
    #     # get stars in tile by reading npys that are spaced at 30 x 10 degrees in RA/Dec
    #     # star_db = self.app.get_path('star_db')
    #     min_ra, max_ra = tile['min_ra'], tile['max_ra']
    #     min_dec, max_dec = tile['min_dec'], tile['max_dec']
    #     dec_tiles = np.arange(dec2tile(min_dec), dec2tile(max_dec) + DEC_STEP, DEC_STEP)
    #     if min_ra < max_ra:
    #         ra_tiles = np.arange(ra2tile(min_ra), ra2tile(max_ra) + RA_STEP, RA_STEP)
    #     else:
    #         ra_tiles = list(range(ra2tile(min_ra), 360, RA_STEP)) + list(
    #             range(0, ra2tile(max_ra) + RA_STEP, 30)
    #         )

    #     # here we could cache when implemented as Joc component
    #     ras, decs, mags = np.array([]), np.array([]), np.array([])
    #     for ra in ra_tiles:
    #         for dec in dec_tiles:
    #             #dat = np.load(os.path.join(star_db, 'r{:}_d{:}.npz'.format(ra, dec)))
    #             quad = 'r{:}_d{:}_'.format(ra, dec)
    #             ras = np.append(ras, self.db[quad+'ra'])
    #             decs = np.append(decs, self.db[quad+'dec'])
    #             mags = np.append(mags, self.db[quad+'mag']/100.) # mags are ints * 100

    #     # restrict to tile
    #     if min_ra < max_ra:
    #         locs = (
    #             (ras >= min_ra) & (ras < max_ra) & (decs >= min_dec) & (decs < max_dec)
    #         )
    #     else:
    #         locs = (
    #             ((ras >= min_ra) | (ras < max_ra))
    #             & (decs >= min_dec)
    #             & (decs < max_dec)
    #         )
    #     return ras[locs], decs[locs], mags[locs]

    def degrees_per_pixel(self):
        arcsec_pp = (
            self.binning
            * (self.pixel_height * 1e-6 / (self.focal_length / 1000))
            * (206265)
        )
        #  self.info('{:4.2f}"/pixel'.format(arcsec_pp))
        return arcsec_pp / 3600



''' Various astro calcs mainly based on Meuss. 
'''

import numpy as np
import math
import time
from datetime import datetime

def julian_date(when):
	# from Meuss p 61; 'when' is a datetime object

	y = when.year
	m = when.month
	d = when.day + when.hour/24 + when.minute/(24*60) + when.second/(24*3600)

	if m < 3:
		y -= 1
		m += 12

	a = int(y / 100)

	if y >= 1582 and m >= 10:
		# Gregorian
		a = int(y/100)
		b = 2 - a + int(a / 4)
	else: 
		# Julian
		b = 0

	jd = int(365.25 * (y + 4716)) + int(30.6001 * (m + 1)) + d + b - 1524.5
	return jd


def to_range(x, d):
	# reduce x to range 0-d by adding or subtracting multiples of d
	if x < 0:
		return x - int((x / d) - 1) * d
	else:
		return x - int((x / d)) * d

def local_sidereal_time(when, longitude):
	# direct method of Meuss p87

	# when must be in UT
	jd = julian_date(when)
	t = (jd - 2451545.0) / 36525.0
	mst = 280.46061837 + 360.98564736629 * (jd - 2451545.0) + .000387933 * t**2 - t**3 / 38710000

	# convert to 0-360
	mst = to_range(mst, 360)

	# convert from Greenwich to local
	lst = mst + longitude

	return lst

def sun_altitude(when, latitude, longitude):
	# Meuss p163+

	jd = julian_date(when)
	rads = math.pi / 180.


	t = (jd - 2451545.0) / 36525.0
	L0 = 280.46646 + 36000.76983 * t + 0.0003032 * t * t
	L0 = to_range(L0, 360)
	M = 357.52911 + 35999.05029 * t - 0.0001537 * t * t
	#e = 0.016708634 - 0.000042037 * t - 0.0000001267 * t * t
	C = (1.914602 - 0.004817 * t - 0.000014 * t * t) * np.sin(M * rads) + \
		(0.019993 - 0.000101 * t) * np.sin(2 * M * rads) + \
		0.000289 * np.sin(3 * M * rads)
	long_sun = L0 + C
	#v = M + C
	# R = (1.000001018 * (1 - e * e)) / (1 + e * np.cos(v * rads))
	sigma = 125.04 - 1934.136 * t
	lam = long_sun - 0.00569 - 0.00478 * np.sin(sigma * rads)
	ep = 23 + (26/60) + (21.448/3600) - (46.815*t + 0.00059 * t**2 - 0.001813*t**3) / 3600
	ep_corr = ep + 0.00256 * np.cos(sigma * rads)
	ra = np.arctan2(np.cos(ep_corr * rads) * np.sin(lam * rads), np.cos(lam * rads)) / rads
	ra = to_range(ra, 360)
	dec = np.arcsin(np.sin(ep_corr * rads) * np.sin(lam * rads)) / rads

	# now convert to locale

	ts = time.time()
	utc_offset = (datetime.fromtimestamp(ts) - datetime.utcfromtimestamp(ts)).total_seconds() / 3600.0
	lst = local_sidereal_time(when, longitude)
	lat = latitude * rads
	H = (-utc_offset*15 + lst - ra) * rads
	alt = np.arcsin(np.sin(lat) * np.sin(dec * rads) + np.cos(lat) * np.cos(dec * rads) * np.cos(H)) / rads

	return alt

